use image::{imageops::FilterType, io::Reader as ImageReader, ImageOutputFormat};
use std::{cmp::max, fs, io::Cursor, path::PathBuf, sync::mpsc::Receiver};

const IMG_MAX_SIZE: u32 = 2048;
const JPEG_QUALITY: u8 = 100; // let ssdv do the compression

#[derive(Debug)]
struct ImgInfo {
    path: PathBuf,
    path_idx: usize,
    size_bytes: u64,
    txed: bool,
    found_on_iter: u64,
}

impl ImgInfo {
    pub fn compressed_jpeg_bytes(&self, dim: u32) -> anyhow::Result<Vec<u8>> {
        let mut img = ImageReader::open(&self.path)?.decode()?;

        // want an image that's $dim on one side
        // resize so that the larger dimension is $dim
        let (mut w, mut h) = if img.width() > dim || img.height() > dim {
            let (w, h) = resize_dimensions(img.width(), img.height(), dim, dim, false);

            img = img.resize_exact(w, h, FilterType::Triangle);

            (w, h)
        } else {
            (img.width(), img.height())
        };

        // make both dimensions divisible by 16 (SSDV requirement)
        let wdiff = w % 16;
        let hdiff = h % 16;
        w -= wdiff;
        h -= hdiff;

        let img = img.crop_imm(wdiff / 2, hdiff / 2, w, h);

        let mut bytes = vec![];

        let img = img.to_rgb8();
        img.write_to(
            &mut Cursor::new(&mut bytes),
            ImageOutputFormat::Jpeg(JPEG_QUALITY),
        )?;

        Ok(bytes)
    }

    pub fn lossless_jpeg_bytes(&self) -> anyhow::Result<Vec<u8>> {
        let mut img = ImageReader::open(&self.path)?.decode()?;

        if img.width() > IMG_MAX_SIZE || img.height() > IMG_MAX_SIZE {
            img = img.resize(IMG_MAX_SIZE, IMG_MAX_SIZE, FilterType::Lanczos3);
        };

        // make both dimensions divisible by 16 (SSDV requirement)
        let mut w = img.width();
        let mut h = img.height();
        let wdiff = w % 16;
        let hdiff = h % 16;
        w -= wdiff;
        h -= hdiff;

        let img = img.crop_imm(wdiff / 2, hdiff / 2, w, h);

        let mut bytes = vec![];

        let img = img.to_rgb8();
        img.write_to(
            &mut Cursor::new(&mut bytes),
            ImageOutputFormat::Jpeg(JPEG_QUALITY),
        )?;

        Ok(bytes)
    }
}

pub struct ImgManager {
    paths: Vec<PathBuf>,
    imgs: Vec<ImgInfo>,
    // used to receive image requests, which will be sent at full resolution
    rx: Receiver<u8>,

    // state for iterator
    iter: u64,
    path_idx: usize,
    found: bool,
    tx_idx: u8,
    // map between sent image ids, and the images
    // in the array above
    // used for decoding rx requests
    img_map: [usize; 256],
}

impl ImgManager {
    pub fn new(paths: Vec<PathBuf>, rx: Receiver<u8>) -> Self {
        Self {
            paths,
            imgs: vec![],
            rx,

            iter: 0,
            path_idx: 0,
            found: false,
            tx_idx: 0,
            img_map: [0; 256],
        }
    }

    // not making this an iterator since
    // it's entirely valid to have a Some(...)
    // after a None
    pub fn next(&mut self, dim: Option<u32>) -> Option<(u8, Vec<u8>)> {
        loop {
            if let Ok(req) = self.rx.try_recv() {
                let img = self
                    .img_map
                    .get(usize::from(req))
                    .and_then(|id| self.imgs.get(*id))
                    .and_then(|img| img.lossless_jpeg_bytes().ok());

                if let Some(img) = img {
                    return Some((self.inc_tx(None), img));
                }
            }

            if let Err(e) = self.scan_path() {
                eprintln!(
                    "Error scanning {}: {}",
                    self.paths[self.path_idx].display(),
                    e
                );
            }

            // new images
            let new = self
                .imgs
                .iter_mut()
                .enumerate()
                .filter(|(_, img)| !img.txed) // non-sent images only
                .filter(|(_, img)| img.path_idx == self.path_idx)
                .filter(|(_, img)| img.found_on_iter == self.iter)
                .max_by_key(|(_, img)| img.size_bytes);

            if let Some((new_id, new)) = new {
                new.txed = true;
                let bytes = match dim {
                    Some(d) => new.compressed_jpeg_bytes(d),
                    None => new.lossless_jpeg_bytes(),
                };

                self.inc_iter(true);

                // not perfect. ideally we would keep looking through the new images
                // if this Err'd
                if let Ok(x) = bytes {
                    return Some((self.inc_tx(Some(new_id)), x));
                }
            }

            // old images
            let old = self
                .imgs
                .iter_mut()
                .enumerate()
                .filter(|(_, img)| !img.txed) // non-sent images only
                .filter(|(_, img)| img.path_idx == self.path_idx)
                .max_by_key(|(_, img)| img.size_bytes);

            if let Some((old_idx, old)) = old {
                old.txed = true;
                let bytes = match dim {
                    Some(d) => old.compressed_jpeg_bytes(d),
                    None => old.lossless_jpeg_bytes(),
                };

                self.inc_iter(true);

                if let Ok(x) = bytes {
                    return Some((self.inc_tx(Some(old_idx)), x));
                }
            }

            // we got all the way through an iteration without finding anything
            // give up for now
            if !self.inc_iter(false) {
                return None;
            }
        }
    }

    fn inc_iter(&mut self, found: bool) -> bool {
        self.path_idx += 1;

        self.found = self.found || found;

        if self.path_idx == self.paths.len() {
            self.path_idx = 0;
            self.iter += 1;

            let found = self.found;
            self.found = false;
            return found;
        }

        true
    }

    fn inc_tx(&mut self, id: Option<usize>) -> u8 {
        let idx = self.tx_idx;
        self.tx_idx = self.tx_idx.wrapping_add(1);

        if let Some(id) = id {
            self.img_map[usize::from(idx)] = id;
        }

        idx
    }

    fn scan_path(&mut self) -> anyhow::Result<()> {
        // imgs we don't already have
        let new_imgs: Vec<_> = fs::read_dir(&self.paths[self.path_idx])?
            // only entries we can actually read without any errors
            .filter_map(|p| {
                p.ok()
                    .and_then(|pp| pp.metadata().ok().map(|meta| (pp.path(), meta)))
            })
            // only files
            .filter(|(p, _)| p.is_file())
            // only new files
            .filter(|(p, _)| !self.imgs.iter().any(|i| i.path == p.as_path()))
            .collect();

        for (path, metadata) in new_imgs {
            let size_bytes = metadata.len();

            self.imgs.push(ImgInfo {
                path,
                path_idx: self.path_idx,
                size_bytes,
                txed: false,
                found_on_iter: self.iter,
            });
        }

        Ok(())
    }
}

// shamelessly stolen from the image crate
fn resize_dimensions(width: u32, height: u32, nwidth: u32, nheight: u32, fill: bool) -> (u32, u32) {
    let wratio = nwidth as f64 / width as f64;
    let hratio = nheight as f64 / height as f64;

    let ratio = if fill {
        f64::max(wratio, hratio)
    } else {
        f64::min(wratio, hratio)
    };

    let nw = max((width as f64 * ratio).round() as u64, 1);
    let nh = max((height as f64 * ratio).round() as u64, 1);

    if nw > u64::from(u32::MAX) {
        let ratio = u32::MAX as f64 / width as f64;
        (u32::MAX, max((height as f64 * ratio).round() as u32, 1))
    } else if nh > u64::from(u32::MAX) {
        let ratio = u32::MAX as f64 / height as f64;
        (max((width as f64 * ratio).round() as u32, 1), u32::MAX)
    } else {
        (nw as u32, nh as u32)
    }
}
